@@ -8,9 +8,11 @@ from twisted.python import log
 from ..wormhole import wormhole
 from ..transit import TransitReceiver
 from ..errors import TransferError, WormholeClosedError
-from ..util import dict_to_bytes, bytes_to_dict, bytes_to_hexstr
+from ..util import (dict_to_bytes, bytes_to_dict, bytes_to_hexstr,
+                    estimate_free_space)
 
 APPID = u"lothar.com/wormhole/text-or-file-xfer"
+VERIFY_TIMER = 1
 
 class RespondError(Exception):
     def __init__(self, response):
@@ -40,7 +42,7 @@ class TwistedReceiver:
         self._transit_receiver = None
 
     def _msg(self, *args, **kwargs):
-        print(*args, file=self.args.stdout, **kwargs)
+        print(*args, file=self.args.stderr, **kwargs)
 
     @inlineCallbacks
     def go(self):
@@ -55,8 +57,8 @@ class TwistedReceiver:
             # with the user handing off the wormhole code
             yield self._tor_manager.start()
 
-        w = wormhole(APPID, self.args.relay_url, self._reactor,
-                     self._tor_manager, timing=self.args.timing)
+        w = wormhole(self.args.appid or APPID, self.args.relay_url,
+                     self._reactor, self._tor_manager, timing=self.args.timing)
         # I wanted to do this instead:
         #
         #    try:
@@ -74,7 +76,16 @@ class TwistedReceiver:
     @inlineCallbacks
     def _go(self, w):
         yield self._handle_code(w)
-        verifier = yield w.verify()
+        yield w.establish_key()
+        def on_slow_connection():
+            print(u"Key established, waiting for confirmation...",
+                  file=self.args.stderr)
+        notify = self._reactor.callLater(VERIFY_TIMER, on_slow_connection)
+        try:
+            verifier = yield w.verify()
+        finally:
+            if not notify.called:
+                notify.cancel()
         self._show_verifier(verifier)
 
         want_offer = True
@@ -190,7 +201,7 @@ class TwistedReceiver:
 
     def _handle_text(self, them_d, w):
         # we're receiving a text message
-        self._msg(them_d["message"])
+        print(them_d["message"], file=self.args.stdout)
         self._send_data({"answer": {"message_ack": "ok"}}, w)
 
     def _handle_file(self, them_d):
@@ -198,6 +209,11 @@ class TwistedReceiver:
         self.abs_destname = self._decide_destname("file",
                                                   file_data["filename"])
         self.xfersize = file_data["filesize"]
+        free = estimate_free_space(self.abs_destname)
+        if free is not None and free < self.xfersize:
+            self._msg(u"Error: insufficient free space (%sB) for file (%sB)"
+                      % (free, self.xfersize))
+            raise TransferRejectedError()
 
         self._msg(u"Receiving file (%s) into: %s" %
                   (naturalsize(self.xfersize), os.path.basename(self.abs_destname)))
@@ -214,6 +230,11 @@ class TwistedReceiver:
         self.abs_destname = self._decide_destname("directory",
                                                   file_data["dirname"])
         self.xfersize = file_data["zipsize"]
+        free = estimate_free_space(self.abs_destname)
+        if free is not None and free < file_data["numbytes"]:
+            self._msg(u"Error: insufficient free space (%sB) for directory (%sB)"
+                      % (free, file_data["numbytes"]))
+            raise TransferRejectedError()
 
         self._msg(u"Receiving directory (%s) into: %s/" %
                   (naturalsize(self.xfersize), os.path.basename(self.abs_destname)))
@@ -232,8 +253,7 @@ class TwistedReceiver:
 
         # get confirmation from the user before writing to the local directory
         if os.path.exists(abs_destname):
-            self._msg(u"Error: refusing to overwrite existing %s %s" %
-                      (mode, destname))
+            self._msg(u"Error: refusing to overwrite existing '%s'" % destname)
             raise TransferRejectedError()
         return abs_destname
 
@@ -263,7 +283,7 @@ class TwistedReceiver:
         self._msg(u"Receiving (%s).." % record_pipe.describe())
 
         with self.args.timing.add("rx file"):
-            progress = tqdm(file=self.args.stdout,
+            progress = tqdm(file=self.args.stderr,
                             disable=self.args.hide_progress,
                             unit="B", unit_scale=True, total=self.xfersize)
             hasher = hashlib.sha256()
