@@ -225,7 +225,7 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
     @inlineCallbacks
     def _do_test(self, as_subprocess=False,
                  mode="text", addslash=False, override_filename=False):
-        assert mode in ("text", "file", "directory")
+        assert mode in ("text", "file", "directory", "slow-text")
         send_cfg = config("send")
         recv_cfg = config("receive")
         message = "blah blah blah ponies"
@@ -244,7 +244,7 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
         receive_dir = self.mktemp()
         os.mkdir(receive_dir)
 
-        if mode == "text":
+        if mode == "text" or mode == "slow-text":
             send_cfg.text = message
 
         elif mode == "file":
@@ -348,8 +348,14 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
 
             # The sender might fail, leaving the receiver hanging, or vice
             # versa. Make sure we don't wait on one side exclusively
+            if mode == "slow-text":
+                with mock.patch.object(cmd_send, "VERIFY_TIMER", 0), \
+                      mock.patch.object(cmd_receive, "VERIFY_TIMER", 0):
+                    yield gatherResults([send_d, receive_d], True)
+            else:
+                yield gatherResults([send_d, receive_d], True)
+
 
-            yield gatherResults([send_d, receive_d], True)
             send_stdout = send_cfg.stdout.getvalue()
             send_stderr = send_cfg.stderr.getvalue()
             receive_stdout = recv_cfg.stdout.getvalue()
@@ -361,64 +367,70 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
 
         self.maxDiff = None # show full output for assertion failures
 
-        self.failUnlessEqual(send_stderr, "",
-                             (send_stdout, send_stderr))
-        self.failUnlessEqual(receive_stderr, "",
-                             (receive_stdout, receive_stderr))
+        key_established = ""
+        if mode == "slow-text":
+            key_established = "Key established, waiting for confirmation...\n"
+
+        self.assertEqual(send_stdout, "")
 
         # check sender
-        if mode == "text":
+        if mode == "text" or mode == "slow-text":
             expected = ("Sending text message ({bytes:d} Bytes){NL}"
                         "On the other computer, please run: "
                         "wormhole receive{NL}"
                         "Wormhole code is: {code}{NL}{NL}"
+                        "{KE}"
                         "text message sent{NL}").format(bytes=len(message),
                                                         code=send_cfg.code,
-                                                        NL=NL)
-            self.failUnlessEqual(send_stdout, expected)
+                                                        NL=NL,
+                                                        KE=key_established)
+            self.failUnlessEqual(send_stderr, expected)
         elif mode == "file":
             self.failUnlessIn("Sending {size:s} file named '{name}'{NL}"
                               .format(size=naturalsize(len(message)),
                                       name=send_filename,
-                                      NL=NL), send_stdout)
+                                      NL=NL), send_stderr)
             self.failUnlessIn("On the other computer, please run: "
                               "wormhole receive{NL}"
                               "Wormhole code is: {code}{NL}{NL}"
                               .format(code=send_cfg.code, NL=NL),
-                              send_stdout)
+                              send_stderr)
             self.failUnlessIn("File sent.. waiting for confirmation{NL}"
                               "Confirmation received. Transfer complete.{NL}"
-                              .format(NL=NL), send_stdout)
+                              .format(NL=NL), send_stderr)
         elif mode == "directory":
-            self.failUnlessIn("Sending directory", send_stdout)
-            self.failUnlessIn("named 'testdir'", send_stdout)
+            self.failUnlessIn("Sending directory", send_stderr)
+            self.failUnlessIn("named 'testdir'", send_stderr)
             self.failUnlessIn("On the other computer, please run: "
                               "wormhole receive{NL}"
                               "Wormhole code is: {code}{NL}{NL}"
-                              .format(code=send_cfg.code, NL=NL), send_stdout)
+                              .format(code=send_cfg.code, NL=NL), send_stderr)
             self.failUnlessIn("File sent.. waiting for confirmation{NL}"
                               "Confirmation received. Transfer complete.{NL}"
-                              .format(NL=NL), send_stdout)
+                              .format(NL=NL), send_stderr)
 
         # check receiver
-        if mode == "text":
-            self.failUnlessEqual(receive_stdout, message+NL)
+        if mode == "text" or mode == "slow-text":
+            self.assertEqual(receive_stdout, message+NL)
+            self.assertEqual(receive_stderr, key_established)
         elif mode == "file":
+            self.failUnlessEqual(receive_stdout, "")
             self.failUnlessIn("Receiving file ({size:s}) into: {name}"
                               .format(size=naturalsize(len(message)),
-                                      name=receive_filename), receive_stdout)
-            self.failUnlessIn("Received file written to ", receive_stdout)
+                                      name=receive_filename), receive_stderr)
+            self.failUnlessIn("Received file written to ", receive_stderr)
             fn = os.path.join(receive_dir, receive_filename)
             self.failUnless(os.path.exists(fn))
             with open(fn, "r") as f:
                 self.failUnlessEqual(f.read(), message)
         elif mode == "directory":
+            self.failUnlessEqual(receive_stdout, "")
             want = (r"Receiving directory \(\d+ \w+\) into: {name}/"
                     .format(name=receive_dirname))
-            self.failUnless(re.search(want, receive_stdout),
-                            (want, receive_stdout))
+            self.failUnless(re.search(want, receive_stderr),
+                            (want, receive_stderr))
             self.failUnlessIn("Received files written to {name}"
-                              .format(name=receive_dirname), receive_stdout)
+                              .format(name=receive_dirname), receive_stderr)
             fn = os.path.join(receive_dir, receive_dirname)
             self.failUnless(os.path.exists(fn), fn)
             for i in range(5):
@@ -448,8 +460,13 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
     def test_directory_override(self):
         return self._do_test(mode="directory", override_filename=True)
 
+    def test_slow_text(self):
+        return self._do_test(mode="slow-text")
+
     @inlineCallbacks
-    def test_file_noclobber(self):
+    def _do_test_fail(self, mode, failmode):
+        assert mode in ("file", "directory")
+        assert failmode in ("noclobber", "toobig")
         send_cfg = config("send")
         recv_cfg = config("receive")
 
@@ -458,29 +475,46 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
             cfg.relay_url = self.relayurl
             cfg.transit_helper = ""
             cfg.listen = False
-            cfg.code = code = "1-abc"
+            cfg.code = "1-abc"
             cfg.stdout = io.StringIO()
             cfg.stderr = io.StringIO()
 
-        message = "test message"
-
-        recv_cfg.accept_file = True
-
         send_dir = self.mktemp()
         os.mkdir(send_dir)
         receive_dir = self.mktemp()
         os.mkdir(receive_dir)
+        recv_cfg.accept_file = True # don't ask for permission
 
-        send_filename = "testfile"
-        with open(os.path.join(send_dir, send_filename), "w") as f:
-            f.write(message)
-        send_cfg.what = receive_filename = send_filename
-        recv_cfg.what = receive_filename
+        if mode == "file":
+            message = "test message\n"
+            send_cfg.what = receive_name = send_filename = "testfile"
+            fn = os.path.join(send_dir, send_filename)
+            with open(fn, "w") as f:
+                f.write(message)
+            size = os.stat(fn).st_size
 
-        PRESERVE = "don't clobber me\n"
-        clobberable = os.path.join(receive_dir, receive_filename)
-        with open(clobberable, "w") as f:
-            f.write(PRESERVE)
+        elif mode == "directory":
+            # $send_dir/
+            # $send_dir/$dirname/
+            # $send_dir/$dirname/[12345]
+            # cd $send_dir && wormhole send $dirname
+            # cd $receive_dir && wormhole receive
+            # expect: $receive_dir/$dirname/[12345]
+
+            size = 0
+            send_cfg.what = receive_name = send_dirname = "testdir"
+            os.mkdir(os.path.join(send_dir, send_dirname))
+            for i in range(5):
+                path = os.path.join(send_dir, send_dirname, str(i))
+                with open(path, "w") as f:
+                    f.write("test message %d\n" % i)
+                size += os.stat(path).st_size
+
+        if failmode == "noclobber":
+            PRESERVE = "don't clobber me\n"
+            clobberable = os.path.join(receive_dir, receive_name)
+            with open(clobberable, "w") as f:
+                f.write(PRESERVE)
 
         send_cfg.cwd = send_dir
         send_d = cmd_send.send(send_cfg)
@@ -488,13 +522,17 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
         recv_cfg.cwd = receive_dir
         receive_d = cmd_receive.receive(recv_cfg)
 
-        # both sides will fail because of the pre-existing file
-
-        f = yield self.assertFailure(send_d, TransferError)
-        self.assertEqual(str(f), "remote error, transfer abandoned: transfer rejected")
-
-        f = yield self.assertFailure(receive_d, TransferError)
-        self.assertEqual(str(f), "transfer rejected")
+        # both sides will fail
+        if failmode == "noclobber":
+            free_space = 10000000
+        else:
+            free_space = 0
+        with mock.patch("wormhole.cli.cmd_receive.estimate_free_space",
+                        return_value=free_space):
+            f = yield self.assertFailure(send_d, TransferError)
+            self.assertEqual(str(f), "remote error, transfer abandoned: transfer rejected")
+            f = yield self.assertFailure(receive_d, TransferError)
+            self.assertEqual(str(f), "transfer rejected")
 
         send_stdout = send_cfg.stdout.getvalue()
         send_stderr = send_cfg.stderr.getvalue()
@@ -507,34 +545,78 @@ class PregeneratedCode(ServerBase, ScriptsBase, unittest.TestCase):
 
         self.maxDiff = None # show full output for assertion failures
 
-        self.failUnlessEqual(send_stderr, "",
-                             (send_stdout, send_stderr))
-        self.failUnlessEqual(receive_stderr, "",
-                             (receive_stdout, receive_stderr))
+        self.assertEqual(send_stdout, "")
+        self.assertEqual(receive_stdout, "")
 
         # check sender
-        self.failUnlessIn("Sending {size:s} file named '{name}'{NL}"
-                          .format(size=naturalsize(len(message)),
-                                  name=send_filename,
-                                  NL=NL), send_stdout)
-        self.failUnlessIn("On the other computer, please run: "
-                          "wormhole receive{NL}"
-                          "Wormhole code is: {code}{NL}{NL}"
-                          .format(code=code, NL=NL),
-                          send_stdout)
-        self.failIfIn("File sent.. waiting for confirmation{NL}"
-                      "Confirmation received. Transfer complete.{NL}"
-                      .format(NL=NL), send_stdout)
+        if mode == "file":
+            self.failUnlessIn("Sending {size:s} file named '{name}'{NL}"
+                              .format(size=naturalsize(size),
+                                      name=send_filename,
+                                      NL=NL), send_stderr)
+            self.failUnlessIn("On the other computer, please run: "
+                              "wormhole receive{NL}"
+                              "Wormhole code is: {code}{NL}{NL}"
+                              .format(code=send_cfg.code, NL=NL),
+                              send_stderr)
+            self.failIfIn("File sent.. waiting for confirmation{NL}"
+                          "Confirmation received. Transfer complete.{NL}"
+                          .format(NL=NL), send_stderr)
+        elif mode == "directory":
+            self.failUnlessIn("Sending directory", send_stderr)
+            self.failUnlessIn("named 'testdir'", send_stderr)
+            self.failUnlessIn("On the other computer, please run: "
+                              "wormhole receive{NL}"
+                              "Wormhole code is: {code}{NL}{NL}"
+                              .format(code=send_cfg.code, NL=NL), send_stderr)
+            self.failIfIn("File sent.. waiting for confirmation{NL}"
+                          "Confirmation received. Transfer complete.{NL}"
+                          .format(NL=NL), send_stderr)
 
         # check receiver
-        self.failUnlessIn("Error: "
-                          "refusing to overwrite existing file testfile{NL}"
-                          .format(NL=NL), receive_stdout)
-        self.failIfIn("Received file written to ", receive_stdout)
-        fn = os.path.join(receive_dir, receive_filename)
-        self.failUnless(os.path.exists(fn))
-        with open(fn, "r") as f:
-            self.failUnlessEqual(f.read(), PRESERVE)
+        if mode == "file":
+            self.failIfIn("Received file written to ", receive_stderr)
+            if failmode == "noclobber":
+                self.failUnlessIn("Error: "
+                                  "refusing to overwrite existing 'testfile'{NL}"
+                                  .format(NL=NL), receive_stderr)
+            else:
+                self.failUnlessIn("Error: "
+                                  "insufficient free space (0B) for file ({size:d}B){NL}"
+                                  .format(NL=NL, size=size), receive_stderr)
+        elif mode == "directory":
+            self.failIfIn("Received files written to {name}"
+                          .format(name=receive_name), receive_stderr)
+            #want = (r"Receiving directory \(\d+ \w+\) into: {name}/"
+            #        .format(name=receive_name))
+            #self.failUnless(re.search(want, receive_stderr),
+            #                (want, receive_stderr))
+            if failmode == "noclobber":
+                self.failUnlessIn("Error: "
+                                  "refusing to overwrite existing 'testdir'{NL}"
+                                  .format(NL=NL), receive_stderr)
+            else:
+                self.failUnlessIn("Error: "
+                                  "insufficient free space (0B) for directory ({size:d}B){NL}"
+                                  .format(NL=NL, size=size), receive_stderr)
+
+        if failmode == "noclobber":
+            fn = os.path.join(receive_dir, receive_name)
+            self.failUnless(os.path.exists(fn))
+            with open(fn, "r") as f:
+                self.failUnlessEqual(f.read(), PRESERVE)
+
+        # check server stats
+        self._rendezvous.get_stats()
+
+    def test_fail_file_noclobber(self):
+        return self._do_test_fail("file", "noclobber")
+    def test_fail_directory_noclobber(self):
+        return self._do_test_fail("directory", "noclobber")
+    def test_fail_file_toobig(self):
+        return self._do_test_fail("file", "toobig")
+    def test_fail_directory_toobig(self):
+        return self._do_test_fail("directory", "toobig")
 
 class NotWelcome(ServerBase, unittest.TestCase):
     def setUp(self):
@@ -654,3 +736,34 @@ class ExtractFile(unittest.TestCase):
         zi.filename = "/etc/passwd"
         e = self.assertRaises(ValueError, ef, zf, zi, extract_dir)
         self.assertIn("malicious zipfile", str(e))
+
+class AppID(ServerBase, unittest.TestCase):
+    def setUp(self):
+        d = super(AppID, self).setUp()
+        self.cfg = cfg = config("send")
+        # common options for all tests in this suite
+        cfg.hide_progress = True
+        cfg.relay_url = self.relayurl
+        cfg.transit_helper = ""
+        cfg.stdout = io.StringIO()
+        cfg.stderr = io.StringIO()
+        return d
+
+    @inlineCallbacks
+    def test_override(self):
+        # make sure we use the overridden appid, not the default
+        self.cfg.text = "hello"
+        self.cfg.appid = "appid2"
+        self.cfg.code = "1-abc"
+
+        send_d = cmd_send.send(self.cfg)
+        receive_d = cmd_receive.receive(self.cfg)
+
+        yield send_d
+        yield receive_d
+
+        used = self._rendezvous._db.execute("SELECT DISTINCT `app_id`"
+                                            " FROM `nameplate_usage`"
+                                            ).fetchall()
+        self.assertEqual(len(used), 1, used)
+        self.assertEqual(used[0]["app_id"], u"appid2")
