@@ -1,21 +1,47 @@
 from __future__ import print_function, unicode_literals
 import os, json, itertools, time
-from binascii import hexlify
 import mock
 from twisted.trial import unittest
 from twisted.python import log
-from twisted.internet import protocol, reactor, defer
+from twisted.internet import reactor, defer
 from twisted.internet.defer import inlineCallbacks, returnValue
-from twisted.internet.endpoints import clientFromString, connectProtocol
-from twisted.web import client
 from autobahn.twisted import websocket
 from .. import __version__
 from .common import ServerBase
-from ..server import server, rendezvous, transit_server
+from ..server import server, rendezvous
 from ..server.rendezvous import Usage, SidedMessage
 from ..server.database import get_db
 
-class Server(ServerBase, unittest.TestCase):
+class _Util:
+    def _nameplate(self, app, name):
+        np_row = app._db.execute("SELECT * FROM `nameplates`"
+                                 " WHERE `app_id`='appid' AND `name`=?",
+                                 (name,)).fetchone()
+        if not np_row:
+            return None, None
+        npid = np_row["id"]
+        side_rows = app._db.execute("SELECT * FROM `nameplate_sides`"
+                                    " WHERE `nameplates_id`=?",
+                                    (npid,)).fetchall()
+        return np_row, side_rows
+
+    def _mailbox(self, app, mailbox_id):
+        mb_row = app._db.execute("SELECT * FROM `mailboxes`"
+                                 " WHERE `app_id`='appid' AND `id`=?",
+                                 (mailbox_id,)).fetchone()
+        if not mb_row:
+            return None, None
+        side_rows = app._db.execute("SELECT * FROM `mailbox_sides`"
+                                    " WHERE `mailbox_id`=?",
+                                    (mailbox_id,)).fetchall()
+        return mb_row, side_rows
+
+    def _messages(self, app):
+        c = app._db.execute("SELECT * FROM `messages`"
+                            " WHERE `app_id`='appid' AND `mailbox_id`='mid'")
+        return c.fetchall()
+
+class Server(_Util, ServerBase, unittest.TestCase):
     def test_apps(self):
         app1 = self._rendezvous.get_app("appid1")
         self.assertIdentical(app1, self._rendezvous.get_app("appid1"))
@@ -48,18 +74,6 @@ class Server(ServerBase, unittest.TestCase):
         biggest = max(nids)
         self.assert_(1000 <= biggest < 1000000, biggest)
 
-    def _nameplate(self, app, name):
-        np_row = app._db.execute("SELECT * FROM `nameplates`"
-                                 " WHERE `app_id`='appid' AND `name`=?",
-                                 (name,)).fetchone()
-        if not np_row:
-            return None, None
-        npid = np_row["id"]
-        side_rows = app._db.execute("SELECT * FROM `nameplate_sides`"
-                                    " WHERE `nameplates_id`=?",
-                                    (npid,)).fetchall()
-        return np_row, side_rows
-
     def test_nameplate(self):
         app = self._rendezvous.get_app("appid")
         name = app.allocate_nameplate("side1", 0)
@@ -151,17 +165,6 @@ class Server(ServerBase, unittest.TestCase):
         self.assertEqual(usage["result"], "crowded")
 
 
-    def _mailbox(self, app, mailbox_id):
-        mb_row = app._db.execute("SELECT * FROM `mailboxes`"
-                                 " WHERE `app_id`='appid' AND `id`=?",
-                                 (mailbox_id,)).fetchone()
-        if not mb_row:
-            return None, None
-        side_rows = app._db.execute("SELECT * FROM `mailbox_sides`"
-                                    " WHERE `mailbox_id`=?",
-                                    (mailbox_id,)).fetchall()
-        return mb_row, side_rows
-
     def test_mailbox(self):
         app = self._rendezvous.get_app("appid")
         mailbox_id = "mid"
@@ -232,11 +235,6 @@ class Server(ServerBase, unittest.TestCase):
         self.assertEqual(usage["total_time"], 7)
         self.assertEqual(usage["result"], "crowded")
 
-    def _messages(self, app):
-        c = app._db.execute("SELECT * FROM `messages`"
-                            " WHERE `app_id`='appid' AND `mailbox_id`='mid'")
-        return c.fetchall()
-
     def test_messages(self):
         app = self._rendezvous.get_app("appid")
         mailbox_id = "mid"
@@ -504,6 +502,9 @@ class WSClient(websocket.WebSocketClientProtocol):
     def next_non_ack(self):
         while True:
             m = yield self.next_event()
+            if isinstance(m, tuple):
+                print("unexpected onClose", m)
+                raise AssertionError("unexpected onClose")
             if m["type"] != "ack":
                 returnValue(m)
 
@@ -602,7 +603,7 @@ class WSClientSync(unittest.TestCase):
 
 
 
-class WebSocketAPI(ServerBase, unittest.TestCase):
+class WebSocketAPI(_Util, ServerBase, unittest.TestCase):
     def setUp(self):
         self._clients = []
         return ServerBase.setUp(self)
@@ -702,18 +703,6 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
             nids.add(n["id"])
         self.assertEqual(nids, set([nameplate_id1, "np2"]))
 
-    def _nameplate(self, app, name):
-        np_row = app._db.execute("SELECT * FROM `nameplates`"
-                                 " WHERE `app_id`='appid' AND `name`=?",
-                                 (name,)).fetchone()
-        if not np_row:
-            return None, None
-        npid = np_row["id"]
-        side_rows = app._db.execute("SELECT * FROM `nameplate_sides`"
-                                    " WHERE `nameplates_id`=?",
-                                    (npid,)).fetchall()
-        return np_row, side_rows
-
     @inlineCallbacks
     def test_allocate(self):
         c1 = yield self.make_client()
@@ -765,6 +754,11 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
         mailbox_id = m["mailbox"]
         self.assertEqual(type(mailbox_id), type(""))
 
+        c1.send("claim", nameplate="np1")
+        err = yield c1.next_non_ack()
+        self.assertEqual(err["type"], "error", err)
+        self.assertEqual(err["error"], "only one claim per connection")
+
         nids = app.get_nameplate_ids()
         self.assertEqual(len(nids), 1)
         self.assertEqual("np1", list(nids)[0])
@@ -807,14 +801,14 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
         err = yield c1.next_non_ack()
         self.assertEqual(err["type"], "error")
         self.assertEqual(err["error"],
-                         "must claim a nameplate before releasing it")
+                         "release without nameplate must follow claim")
 
         c1.send("claim", nameplate="np1")
         yield c1.next_non_ack()
 
         c1.send("release")
         m = yield c1.next_non_ack()
-        self.assertEqual(m["type"], "released")
+        self.assertEqual(m["type"], "released", m)
 
         np_row, side_rows = self._nameplate(app, "np1")
         claims = [(row["side"], row["claimed"]) for row in side_rows]
@@ -824,8 +818,45 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
         c1.send("release") # no longer claimed
         err = yield c1.next_non_ack()
         self.assertEqual(err["type"], "error")
+        self.assertEqual(err["error"], "only one release per connection")
+
+    @inlineCallbacks
+    def test_release_named(self):
+        c1 = yield self.make_client()
+        yield c1.next_non_ack()
+        c1.send("bind", appid="appid", side="side")
+
+        c1.send("claim", nameplate="np1")
+        yield c1.next_non_ack()
+
+        c1.send("release", nameplate="np1")
+        m = yield c1.next_non_ack()
+        self.assertEqual(m["type"], "released", m)
+
+    @inlineCallbacks
+    def test_release_named_ignored(self):
+        c1 = yield self.make_client()
+        yield c1.next_non_ack()
+        c1.send("bind", appid="appid", side="side")
+
+        c1.send("release", nameplate="np1") # didn't do claim first, ignored
+        m = yield c1.next_non_ack()
+        self.assertEqual(m["type"], "released", m)
+
+    @inlineCallbacks
+    def test_release_named_mismatch(self):
+        c1 = yield self.make_client()
+        yield c1.next_non_ack()
+        c1.send("bind", appid="appid", side="side")
+
+        c1.send("claim", nameplate="np1")
+        yield c1.next_non_ack()
+
+        c1.send("release", nameplate="np2") # mismatching nameplate
+        err = yield c1.next_non_ack()
+        self.assertEqual(err["type"], "error")
         self.assertEqual(err["error"],
-                         "must claim a nameplate before releasing it")
+                         "release and claim must use same nameplate")
 
     @inlineCallbacks
     def test_open(self):
@@ -860,7 +891,7 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
         c1.send("open", mailbox="mb1")
         err = yield c1.next_non_ack()
         self.assertEqual(err["type"], "error")
-        self.assertEqual(err["error"], "you already have a mailbox open")
+        self.assertEqual(err["error"], "only one open per connection")
 
     @inlineCallbacks
     def test_add(self):
@@ -907,7 +938,7 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
         c1.send("close", mood="mood") # must open first
         err = yield c1.next_non_ack()
         self.assertEqual(err["type"], "error")
-        self.assertEqual(err["error"], "must open mailbox before closing")
+        self.assertEqual(err["error"], "close without mailbox must follow open")
 
         c1.send("open", mailbox="mb1")
         yield c1.sync()
@@ -921,8 +952,46 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
 
         c1.send("close", mood="mood") # already closed
         err = yield c1.next_non_ack()
+        self.assertEqual(err["type"], "error", m)
+        self.assertEqual(err["error"], "only one close per connection")
+
+    @inlineCallbacks
+    def test_close_named(self):
+        c1 = yield self.make_client()
+        yield c1.next_non_ack()
+        c1.send("bind", appid="appid", side="side")
+
+        c1.send("open", mailbox="mb1")
+        yield c1.sync()
+
+        c1.send("close", mailbox="mb1", mood="mood")
+        m = yield c1.next_non_ack()
+        self.assertEqual(m["type"], "closed")
+
+    @inlineCallbacks
+    def test_close_named_ignored(self):
+        c1 = yield self.make_client()
+        yield c1.next_non_ack()
+        c1.send("bind", appid="appid", side="side")
+
+        c1.send("close", mailbox="mb1", mood="mood") # no open first, ignored
+        m = yield c1.next_non_ack()
+        self.assertEqual(m["type"], "closed")
+
+    @inlineCallbacks
+    def test_close_named_mismatch(self):
+        c1 = yield self.make_client()
+        yield c1.next_non_ack()
+        c1.send("bind", appid="appid", side="side")
+
+        c1.send("open", mailbox="mb1")
+        yield c1.sync()
+
+        c1.send("close", mailbox="mb2", mood="mood")
+        err = yield c1.next_non_ack()
         self.assertEqual(err["type"], "error")
-        self.assertEqual(err["error"], "must open mailbox before closing")
+        self.assertEqual(err["error"], "open and close must use same mailbox")
+
 
     @inlineCallbacks
     def test_disconnect(self):
@@ -945,6 +1014,163 @@ class WebSocketAPI(ServerBase, unittest.TestCase):
             yield d
         self.assertFalse(mb1.has_listeners())
 
+    @inlineCallbacks
+    def test_interrupted_client_nameplate(self):
+        # a client's interactions with the server might be split over
+        # multiple sequential WebSocket connections, e.g. when the server is
+        # bounced and the client reconnects, or vice versa
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        app = self._rendezvous.get_app("appid")
+
+        c.send("claim", nameplate="np1")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "claimed")
+        mailbox_id = m["mailbox"]
+        self.assertEqual(type(mailbox_id), type(""))
+        np_row, side_rows = self._nameplate(app, "np1")
+        claims = [(row["side"], row["claimed"]) for row in side_rows]
+        self.assertEqual(claims, [("side", True)])
+        c.close()
+        yield c.d
+
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        c.send("claim", nameplate="np1") # idempotent
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "claimed")
+        self.assertEqual(m["mailbox"], mailbox_id) # mailbox id is stable
+        np_row, side_rows = self._nameplate(app, "np1")
+        claims = [(row["side"], row["claimed"]) for row in side_rows]
+        self.assertEqual(claims, [("side", True)])
+        c.close()
+        yield c.d
+
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        # we haven't done a claim with this particular connection, but we can
+        # still send a release as long as we include the nameplate
+        c.send("release", nameplate="np1") # release-without-claim
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "released")
+        np_row, side_rows = self._nameplate(app, "np1")
+        self.assertEqual(np_row, None)
+        c.close()
+        yield c.d
+
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        # and the release is idempotent, when done on separate connections
+        c.send("release", nameplate="np1")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "released")
+        np_row, side_rows = self._nameplate(app, "np1")
+        self.assertEqual(np_row, None)
+        c.close()
+        yield c.d
+
+
+    @inlineCallbacks
+    def test_interrupted_client_nameplate_reclaimed(self):
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        app = self._rendezvous.get_app("appid")
+
+        # a new claim on a previously-closed nameplate is forbidden. We make
+        # a new nameplate here and manually open a second claim on it, so the
+        # nameplate stays alive long enough for the code check to happen.
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        c.send("claim", nameplate="np2")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "claimed")
+        app.claim_nameplate("np2", "side2", 0)
+        c.send("release", nameplate="np2")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "released")
+        np_row, side_rows = self._nameplate(app, "np2")
+        claims = sorted([(row["side"], row["claimed"]) for row in side_rows])
+        self.assertEqual(claims, [("side", 0), ("side2", 1)])
+        c.close()
+        yield c.d
+
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        c.send("claim", nameplate="np2") # new claim is forbidden
+        err = yield c.next_non_ack()
+        self.assertEqual(err["type"], "error")
+        self.assertEqual(err["error"], "reclaimed")
+
+        np_row, side_rows = self._nameplate(app, "np2")
+        claims = sorted([(row["side"], row["claimed"]) for row in side_rows])
+        self.assertEqual(claims, [("side", 0), ("side2", 1)])
+        c.close()
+        yield c.d
+
+    @inlineCallbacks
+    def test_interrupted_client_mailbox(self):
+        # a client's interactions with the server might be split over
+        # multiple sequential WebSocket connections, e.g. when the server is
+        # bounced and the client reconnects, or vice versa
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        app = self._rendezvous.get_app("appid")
+        mb1 = app.open_mailbox("mb1", "side2", 0)
+        mb1.add_message(SidedMessage(side="side2", phase="phase",
+                                     body="body", server_rx=0,
+                                     msg_id="msgid"))
+
+        c.send("open", mailbox="mb1")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "message")
+        self.assertEqual(m["body"], "body")
+        self.assertTrue(mb1.has_listeners())
+        c.close()
+        yield c.d
+
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        # open should be idempotent
+        c.send("open", mailbox="mb1")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "message")
+        self.assertEqual(m["body"], "body")
+        mb_row, side_rows = self._mailbox(app, "mb1")
+        openeds = [(row["side"], row["opened"]) for row in side_rows]
+        self.assertIn(("side", 1), openeds) # TODO: why 1, and not True?
+
+        # close on the same connection as open is ok
+        c.send("close", mailbox="mb1", mood="mood")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "closed", m)
+        mb_row, side_rows = self._mailbox(app, "mb1")
+        openeds = [(row["side"], row["opened"]) for row in side_rows]
+        self.assertIn(("side", 0), openeds)
+        c.close()
+        yield c.d
+
+        # close (on a separate connection) is idempotent
+        c = yield self.make_client()
+        yield c.next_non_ack()
+        c.send("bind", appid="appid", side="side")
+        c.send("close", mailbox="mb1", mood="mood")
+        m = yield c.next_non_ack()
+        self.assertEqual(m["type"], "closed", m)
+        mb_row, side_rows = self._mailbox(app, "mb1")
+        openeds = [(row["side"], row["opened"]) for row in side_rows]
+        self.assertIn(("side", 0), openeds)
+        c.close()
+        yield c.d
+
 
 class Summary(unittest.TestCase):
     def test_mailbox(self):
@@ -1066,130 +1292,3 @@ class DumpStats(unittest.TestCase):
         self.assertEqual(data["rendezvous"]["all_time"]["mailboxes_total"], 0)
         self.assertEqual(data["transit"]["all_time"]["total"], 0)
 
-
-class Accumulator(protocol.Protocol):
-    def __init__(self):
-        self.data = b""
-        self.count = 0
-        self._wait = None
-    def waitForBytes(self, more):
-        assert self._wait is None
-        self.count = more
-        self._wait = defer.Deferred()
-        self._check_done()
-        return self._wait
-    def dataReceived(self, data):
-        self.data = self.data + data
-        self._check_done()
-    def _check_done(self):
-        if self._wait and len(self.data) >= self.count:
-            d = self._wait
-            self._wait = None
-            d.callback(self)
-    def connectionLost(self, why):
-        if self._wait:
-            self._wait.errback(RuntimeError("closed"))
-
-class Transit(ServerBase, unittest.TestCase):
-    def test_blur_size(self):
-        blur = transit_server.blur_size
-        self.failUnlessEqual(blur(0), 0)
-        self.failUnlessEqual(blur(1), 10e3)
-        self.failUnlessEqual(blur(10e3), 10e3)
-        self.failUnlessEqual(blur(10e3+1), 20e3)
-        self.failUnlessEqual(blur(15e3), 20e3)
-        self.failUnlessEqual(blur(20e3), 20e3)
-        self.failUnlessEqual(blur(1e6), 1e6)
-        self.failUnlessEqual(blur(1e6+1), 2e6)
-        self.failUnlessEqual(blur(1.5e6), 2e6)
-        self.failUnlessEqual(blur(2e6), 2e6)
-        self.failUnlessEqual(blur(900e6), 900e6)
-        self.failUnlessEqual(blur(1000e6), 1000e6)
-        self.failUnlessEqual(blur(1050e6), 1100e6)
-        self.failUnlessEqual(blur(1100e6), 1100e6)
-        self.failUnlessEqual(blur(1150e6), 1200e6)
-
-    @defer.inlineCallbacks
-    def test_web_request(self):
-        resp = yield client.getPage('http://127.0.0.1:{}/'.format(self.relayport).encode('ascii'))
-        self.assertEqual('Wormhole Relay'.encode('ascii'), resp.strip())
-
-    @defer.inlineCallbacks
-    def test_basic(self):
-        ep = clientFromString(reactor, self.transit)
-        a1 = yield connectProtocol(ep, Accumulator())
-        a2 = yield connectProtocol(ep, Accumulator())
-
-        token1 = b"\x00"*32
-        a1.transport.write(b"please relay " + hexlify(token1) + b"\n")
-        a2.transport.write(b"please relay " + hexlify(token1) + b"\n")
-
-        # a correct handshake yields an ack, after which we can send
-        exp = b"ok\n"
-        yield a1.waitForBytes(len(exp))
-        self.assertEqual(a1.data, exp)
-        s1 = b"data1"
-        a1.transport.write(s1)
-
-        exp = b"ok\n"
-        yield a2.waitForBytes(len(exp))
-        self.assertEqual(a2.data, exp)
-
-        # all data they sent after the handshake should be given to us
-        exp = b"ok\n"+s1
-        yield a2.waitForBytes(len(exp))
-        self.assertEqual(a2.data, exp)
-
-        a1.transport.loseConnection()
-        a2.transport.loseConnection()
-
-    @defer.inlineCallbacks
-    def test_bad_handshake(self):
-        ep = clientFromString(reactor, self.transit)
-        a1 = yield connectProtocol(ep, Accumulator())
-
-        token1 = b"\x00"*32
-        # the server waits for the exact number of bytes in the expected
-        # handshake message. to trigger "bad handshake", we must match.
-        a1.transport.write(b"please DELAY " + hexlify(token1) + b"\n")
-
-        exp = b"bad handshake\n"
-        yield a1.waitForBytes(len(exp))
-        self.assertEqual(a1.data, exp)
-
-        a1.transport.loseConnection()
-
-    @defer.inlineCallbacks
-    def test_binary_handshake(self):
-        ep = clientFromString(reactor, self.transit)
-        a1 = yield connectProtocol(ep, Accumulator())
-
-        binary_bad_handshake = b"\x00\x01\xe0\x0f\n\xff"
-        # the embedded \n makes the server trigger early, before the full
-        # expected handshake length has arrived. A non-wormhole client
-        # writing non-ascii junk to the transit port used to trigger a
-        # UnicodeDecodeError when it tried to coerce the incoming handshake
-        # to unicode, due to the ("\n" in buf) check. This was fixed to use
-        # (b"\n" in buf). This exercises the old failure.
-        a1.transport.write(binary_bad_handshake)
-
-        exp = b"bad handshake\n"
-        yield a1.waitForBytes(len(exp))
-        self.assertEqual(a1.data, exp)
-
-        a1.transport.loseConnection()
-
-    @defer.inlineCallbacks
-    def test_impatience(self):
-        ep = clientFromString(reactor, self.transit)
-        a1 = yield connectProtocol(ep, Accumulator())
-
-        token1 = b"\x00"*32
-        # sending too many bytes is impatience.
-        a1.transport.write(b"please RELAY NOWNOW " + hexlify(token1) + b"\n")
-
-        exp = b"impatient\n"
-        yield a1.waitForBytes(len(exp))
-        self.assertEqual(a1.data, exp)
-
-        a1.transport.loseConnection()
