@@ -1,7 +1,9 @@
 from __future__ import print_function, unicode_literals
 import io
 import gc
+import mock
 from binascii import hexlify, unhexlify
+from collections import namedtuple
 from twisted.trial import unittest
 from twisted.internet import defer, task, endpoints, protocol, address, error
 from twisted.internet.defer import gatherResults, inlineCallbacks
@@ -9,6 +11,8 @@ from twisted.python import log, failure
 from twisted.test import proto_helpers
 from ..errors import InternalError
 from .. import transit
+from ..server import transit_server
+from .common import ServerBase
 from nacl.secret import SecretBox
 from nacl.exceptions import CryptoError
 
@@ -128,14 +132,110 @@ class Misc(unittest.TestCase):
         portno = transit.allocate_tcp_port()
         self.assertIsInstance(portno, int)
 
+UnknownHint = namedtuple("UnknownHint", ["stuff"])
+
 class Hints(unittest.TestCase):
     def test_endpoint_from_hint_obj(self):
         c = transit.Common("")
-        ep = c._endpoint_from_hint_obj(transit.DirectTCPV1Hint("localhost", 1234))
-        self.assertIsInstance(ep, endpoints.HostnameEndpoint)
-        ep = c._endpoint_from_hint_obj("unknown:stuff:yowza:pivlor")
-        self.assertEqual(ep, None)
+        efho = c._endpoint_from_hint_obj
+        self.assertIsInstance(efho(transit.DirectTCPV1Hint("host", 1234, 0.0)),
+                              endpoints.HostnameEndpoint)
+        self.assertEqual(efho("unknown:stuff:yowza:pivlor"), None)
+        # c._tor_manager is currently None
+        self.assertEqual(efho(transit.TorTCPV1Hint("host", "port", 0)), None)
+        c._tor_manager = mock.Mock()
+        def tor_ep(hostname, port):
+            if hostname == "non-public":
+                return None
+            return ("tor_ep", hostname, port)
+        c._tor_manager.get_endpoint_for = mock.Mock(side_effect=tor_ep)
+        self.assertEqual(efho(transit.DirectTCPV1Hint("host", 1234, 0.0)),
+                         ("tor_ep", "host", 1234))
+        self.assertEqual(efho(transit.TorTCPV1Hint("host2.onion", 1234, 0.0)),
+                         ("tor_ep", "host2.onion", 1234))
+        self.assertEqual(efho(transit.DirectTCPV1Hint("non-public", 1234, 0.0)),
+                         None)
+        self.assertEqual(efho(UnknownHint("foo")), None)
 
+    def test_comparable(self):
+        h1 = transit.DirectTCPV1Hint("hostname", "port1", 0.0)
+        h1b = transit.DirectTCPV1Hint("hostname", "port1", 0.0)
+        h2 = transit.DirectTCPV1Hint("hostname", "port2", 0.0)
+        r1 = transit.RelayV1Hint(tuple(sorted([h1, h2])))
+        r2 = transit.RelayV1Hint(tuple(sorted([h2, h1])))
+        r3 = transit.RelayV1Hint(tuple(sorted([h1b, h2])))
+        self.assertEqual(r1, r2)
+        self.assertEqual(r2, r3)
+        self.assertEqual(len(set([r1, r2, r3])), 1)
+
+    def test_parse_tcp_v1_hint(self):
+        c = transit.Common("")
+        p = c._parse_tcp_v1_hint
+        self.assertEqual(p({"type": "unknown"}), None)
+        h = p({"type": "direct-tcp-v1", "hostname": "foo", "port": 1234})
+        self.assertEqual(h, transit.DirectTCPV1Hint("foo", 1234, 0.0))
+        h = p({"type": "direct-tcp-v1", "hostname": "foo", "port": 1234,
+               "priority": 2.5})
+        self.assertEqual(h, transit.DirectTCPV1Hint("foo", 1234, 2.5))
+        h = p({"type": "tor-tcp-v1", "hostname": "foo", "port": 1234})
+        self.assertEqual(h, transit.TorTCPV1Hint("foo", 1234, 0.0))
+        h = p({"type": "tor-tcp-v1", "hostname": "foo", "port": 1234,
+               "priority": 2.5})
+        self.assertEqual(h, transit.TorTCPV1Hint("foo", 1234, 2.5))
+        self.assertEqual(p({"type": "direct-tcp-v1"}),
+                         None) # missing hostname
+        self.assertEqual(p({"type": "direct-tcp-v1", "hostname": 12}),
+                         None) # invalid hostname
+        self.assertEqual(p({"type": "direct-tcp-v1", "hostname": "foo"}),
+                         None) # missing port
+        self.assertEqual(p({"type": "direct-tcp-v1", "hostname": "foo",
+                            "port": "not a number"}),
+                         None) # invalid port
+
+    def test_parse_hint_argv(self):
+        def p(hint):
+            stderr = io.StringIO()
+            value = transit.parse_hint_argv(hint, stderr=stderr)
+            return value, stderr.getvalue()
+        h,stderr = p("tcp:host:1234")
+        self.assertEqual(h, transit.DirectTCPV1Hint("host", 1234, 0.0))
+        self.assertEqual(stderr, "")
+
+        h,stderr = p("tcp:host:1234:priority=2.6")
+        self.assertEqual(h, transit.DirectTCPV1Hint("host", 1234, 2.6))
+        self.assertEqual(stderr, "")
+
+        h,stderr = p("$!@#^")
+        self.assertEqual(h, None)
+        self.assertEqual(stderr, "unparseable hint '$!@#^'\n")
+
+        h,stderr = p("unknown:stuff")
+        self.assertEqual(h, None)
+        self.assertEqual(stderr,
+                         "unknown hint type 'unknown' in 'unknown:stuff'\n")
+
+        h,stderr = p("tcp:just-a-hostname")
+        self.assertEqual(h, None)
+        self.assertEqual(stderr,
+                         "unparseable TCP hint (need more colons) 'tcp:just-a-hostname'\n")
+
+        h,stderr = p("tcp:host:number")
+        self.assertEqual(h, None)
+        self.assertEqual(stderr,
+                         "non-numeric port in TCP hint 'tcp:host:number'\n")
+
+        h,stderr = p("tcp:host:1234:priority=bad")
+        self.assertEqual(h, None)
+        self.assertEqual(stderr,
+                         "non-float priority= in TCP hint 'tcp:host:1234:priority=bad'\n")
+
+    def test_describe_hint_obj(self):
+        d = transit.describe_hint_obj
+        self.assertEqual(d(transit.DirectTCPV1Hint("host", 1234, 0.0)),
+                         "tcp:host:1234")
+        self.assertEqual(d(transit.TorTCPV1Hint("host", 1234, 0.0)),
+                         "tor:host:1234")
+        self.assertEqual(d(UnknownHint("stuff")), str(UnknownHint("stuff")))
 
 class Basic(unittest.TestCase):
     @inlineCallbacks
@@ -145,8 +245,9 @@ class Basic(unittest.TestCase):
         hints = yield c.get_connection_hints()
         self.assertEqual(hints, [{"type": "relay-v1",
                                   "hints": [{"type": "direct-tcp-v1",
-                                            "hostname": "host",
-                                            "port": 1234}],
+                                             "hostname": "host",
+                                             "port": 1234,
+                                             "priority": 0.0}],
                                   }])
         self.assertRaises(InternalError, transit.Common, 123)
 
@@ -162,7 +263,7 @@ class Basic(unittest.TestCase):
         c.add_connection_hints([{"type": "relay-v1",
                                  "hints": [{"type": "unknown"}]}])
         self.assertEqual(c._their_direct_hints, [])
-        self.assertEqual(c._their_relay_hints, [])
+        self.assertEqual(c._our_relay_hints, set())
 
     def test_ignore_localhost_hint(self):
         # this actually starts the listener
@@ -1193,21 +1294,32 @@ class FileConsumer(unittest.TestCase):
         self.assertEqual(f.getvalue(), b"."*99+b"!")
 
 
-DIRECT_HINT = {"type": "direct-tcp-v1",
-               "hostname": "direct", "port": 1234}
-RELAY_HINT = {"type": "relay-v1",
-              "hints": [{"type": "direct-tcp-v1",
-                          "hostname": "relay", "port": 1234}]}
-UNUSABLE_HINT = {"type": "unknown"}
-RELAY_HINT2 = {"type": "relay-v1",
-               "hints": [{"type": "direct-tcp-v1",
-                           "hostname": "relay", "port": 1234},
-                          UNUSABLE_HINT]}
-DIRECT_HINT_INTERNAL = transit.DirectTCPV1Hint("direct", 1234)
-RELAY_HINT_FIRST = transit.DirectTCPV1Hint("relay", 1234)
-RELAY_HINT_INTERNAL = transit.RelayV1Hint([RELAY_HINT_FIRST])
+DIRECT_HINT_JSON = {"type": "direct-tcp-v1",
+                    "hostname": "direct", "port": 1234}
+RELAY_HINT_JSON = {"type": "relay-v1",
+                   "hints": [{"type": "direct-tcp-v1",
+                              "hostname": "relay", "port": 1234}]}
+UNRECOGNIZED_HINT_JSON = {"type": "unknown"}
+UNAVAILABLE_HINT_JSON = {"type": "direct-tcp-v1", # e.g. Tor without txtorcon
+                         "hostname": "unavailable", "port": 1234}
+RELAY_HINT2_JSON = {"type": "relay-v1",
+                    "hints": [{"type": "direct-tcp-v1",
+                               "hostname": "relay", "port": 1234},
+                              UNRECOGNIZED_HINT_JSON]}
+UNAVAILABLE_RELAY_HINT_JSON = {"type": "relay-v1",
+                               "hints": [UNAVAILABLE_HINT_JSON]}
 
 class Transit(unittest.TestCase):
+    def setUp(self):
+        self._connectors = []
+        self._waiters = []
+
+    def _start_connector(self, ep, description, is_relay=False):
+        d = defer.Deferred()
+        self._connectors.append(ep)
+        self._waiters.append(d)
+        return d
+
     @inlineCallbacks
     def test_success_direct(self):
         clock = task.Clock()
@@ -1215,54 +1327,38 @@ class Transit(unittest.TestCase):
         s.set_transit_key(b"key")
         hints = yield s.get_connection_hints() # start the listener
         del hints
-        s.add_connection_hints([DIRECT_HINT, UNUSABLE_HINT])
+        s.add_connection_hints([DIRECT_HINT_JSON, UNRECOGNIZED_HINT_JSON])
 
-        connectors = []
-        def _start_connector(ep, description, is_relay=False):
-            d = defer.Deferred()
-            connectors.append(d)
-            return d
-        s._start_connector = _start_connector
+        s._start_connector = self._start_connector
         d = s.connect()
         results = []
         d.addBoth(results.append)
         self.assertEqual(results, [])
-        self.assertEqual(len(connectors), 1)
-        self.assertIsInstance(connectors[0], defer.Deferred)
+        self.assertEqual(len(self._waiters), 1)
+        self.assertIsInstance(self._waiters[0], defer.Deferred)
 
-        connectors[0].callback("winner")
+        self._waiters[0].callback("winner")
         self.assertEqual(results, ["winner"])
 
     def _endpoint_from_hint_obj(self, hint):
-        if hint == DIRECT_HINT_INTERNAL:
-            return "direct"
-        elif hint == RELAY_HINT_FIRST:
-            return "relay"
-        else:
-            return None
+        if isinstance(hint, transit.DirectTCPV1Hint):
+            if hint.hostname == "unavailable":
+                return None
+            return hint.hostname
+        return None
 
     @inlineCallbacks
     def test_wait_for_relay(self):
         clock = task.Clock()
         s = transit.TransitSender("", reactor=clock, no_listen=True)
         s.set_transit_key(b"key")
-        hints = yield s.get_connection_hints() # start the listener
+        hints = yield s.get_connection_hints()
         del hints
-        s.add_connection_hints([DIRECT_HINT, UNUSABLE_HINT, RELAY_HINT])
-
-        direct_connectors = []
-        relay_connectors = []
+        s.add_connection_hints([DIRECT_HINT_JSON,
+                                UNRECOGNIZED_HINT_JSON,
+                                RELAY_HINT_JSON])
         s._endpoint_from_hint_obj = self._endpoint_from_hint_obj
-        def _start_connector(ep, description, is_relay=False):
-            d = defer.Deferred()
-            if ep == "direct":
-                direct_connectors.append(d)
-            elif ep == "relay":
-                relay_connectors.append(d)
-            else:
-                raise ValueError
-            return d
-        s._start_connector = _start_connector
+        s._start_connector = self._start_connector
 
         d = s.connect()
         results = []
@@ -1270,14 +1366,62 @@ class Transit(unittest.TestCase):
         self.assertEqual(results, [])
         # the direct connectors are tried right away, but the relay
         # connectors are stalled for a few seconds
-        self.assertEqual(len(direct_connectors), 1)
-        self.assertEqual(len(relay_connectors), 0)
+        self.assertEqual(self._connectors, ["direct"])
 
         clock.advance(s.RELAY_DELAY + 1.0)
-        self.assertEqual(len(direct_connectors), 1)
-        self.assertEqual(len(relay_connectors), 1)
+        self.assertEqual(self._connectors, ["direct", "relay"])
 
-        direct_connectors[0].callback("winner")
+        self._waiters[0].callback("winner")
+        self.assertEqual(results, ["winner"])
+
+    @inlineCallbacks
+    def test_priorities(self):
+        clock = task.Clock()
+        s = transit.TransitSender("", reactor=clock, no_listen=True)
+        s.set_transit_key(b"key")
+        hints = yield s.get_connection_hints()
+        del hints
+        s.add_connection_hints([
+            {"type": "relay-v1",
+             "hints": [{"type": "direct-tcp-v1",
+                        "hostname": "relay", "port": 1234}]},
+            {"type": "direct-tcp-v1",
+             "hostname": "direct", "port": 1234},
+            {"type": "relay-v1",
+             "hints": [{"type": "direct-tcp-v1", "priority": 2.0,
+                        "hostname": "relay2", "port": 1234},
+                       {"type": "direct-tcp-v1", "priority": 3.0,
+                        "hostname": "relay3", "port": 1234}]},
+            {"type": "relay-v1",
+             "hints": [{"type": "direct-tcp-v1", "priority": 2.0,
+                        "hostname": "relay4", "port": 1234}]},
+            ])
+        s._endpoint_from_hint_obj = self._endpoint_from_hint_obj
+        s._start_connector = self._start_connector
+
+        d = s.connect()
+        results = []
+        d.addBoth(results.append)
+        self.assertEqual(results, [])
+        # direct connector should be used first, then the priority=3.0 relay,
+        # then the two 2.0 relays, then the (default) 0.0 relay
+
+        self.assertEqual(self._connectors, ["direct"])
+
+        clock.advance(s.RELAY_DELAY + 1.0)
+        self.assertEqual(self._connectors, ["direct", "relay3"])
+
+        clock.advance(s.RELAY_DELAY)
+        self.assertIn(self._connectors,
+                      (["direct", "relay3", "relay2", "relay4"],
+                       ["direct", "relay3", "relay4", "relay2"]))
+
+        clock.advance(s.RELAY_DELAY)
+        self.assertIn(self._connectors,
+                      (["direct", "relay3", "relay2", "relay4", "relay"],
+                       ["direct", "relay3", "relay4", "relay2", "relay"]))
+
+        self._waiters[0].callback("winner")
         self.assertEqual(results, ["winner"])
 
     @inlineCallbacks
@@ -1287,21 +1431,13 @@ class Transit(unittest.TestCase):
         s.set_transit_key(b"key")
         hints = yield s.get_connection_hints() # start the listener
         del hints
-        s.add_connection_hints([UNUSABLE_HINT, RELAY_HINT2])
-
-        direct_connectors = []
-        relay_connectors = []
+        # include hints that can't be turned into an endpoint at runtime
+        s.add_connection_hints([UNRECOGNIZED_HINT_JSON,
+                                UNAVAILABLE_HINT_JSON,
+                                RELAY_HINT2_JSON,
+                                UNAVAILABLE_RELAY_HINT_JSON])
         s._endpoint_from_hint_obj = self._endpoint_from_hint_obj
-        def _start_connector(ep, description, is_relay=False):
-            d = defer.Deferred()
-            if ep == "direct":
-                direct_connectors.append(d)
-            elif ep == "relay":
-                relay_connectors.append(d)
-            else:
-                raise ValueError
-            return d
-        s._start_connector = _start_connector
+        s._start_connector = self._start_connector
 
         d = s.connect()
         results = []
@@ -1309,23 +1445,68 @@ class Transit(unittest.TestCase):
         self.assertEqual(results, [])
         # since there are no usable direct hints, the relay connector will
         # only be stalled for 0 seconds
-        self.assertEqual(len(direct_connectors), 0)
-        self.assertEqual(len(relay_connectors), 0)
+        self.assertEqual(self._connectors, [])
 
         clock.advance(0)
-        self.assertEqual(len(direct_connectors), 0)
-        self.assertEqual(len(relay_connectors), 1)
+        self.assertEqual(self._connectors, ["relay"])
 
-        relay_connectors[0].callback("winner")
+        self._waiters[0].callback("winner")
         self.assertEqual(results, ["winner"])
 
+    @inlineCallbacks
+    def test_no_contenders(self):
+        clock = task.Clock()
+        s = transit.TransitSender("", reactor=clock, no_listen=True)
+        s.set_transit_key(b"key")
+        hints = yield s.get_connection_hints() # start the listener
+        del hints
+        s.add_connection_hints([]) # no hints at all
+        s._endpoint_from_hint_obj = self._endpoint_from_hint_obj
+        s._start_connector = self._start_connector
 
-class Full(unittest.TestCase):
+        d = s.connect()
+        f = self.failureResultOf(d, transit.TransitError)
+        self.assertEqual(str(f.value), "No contenders for connection")
+
+class RelayHandshake(unittest.TestCase):
+    def old_build_relay_handshake(self, key):
+        token = transit.HKDF(key, 32, CTXinfo=b"transit_relay_token")
+        return (token, b"please relay "+hexlify(token)+b"\n")
+
+    def test_old(self):
+        key = b"\x00"
+        token, old_handshake = self.old_build_relay_handshake(key)
+        tc = transit_server.TransitConnection()
+        tc.factory = mock.Mock()
+        tc.factory.connection_got_token = mock.Mock()
+        tc.dataReceived(old_handshake[:-1])
+        self.assertEqual(tc.factory.connection_got_token.mock_calls, [])
+        tc.dataReceived(old_handshake[-1:])
+        self.assertEqual(tc.factory.connection_got_token.mock_calls,
+                         [mock.call(hexlify(token), None, tc)])
+
+    def test_new(self):
+        c = transit.Common(None)
+        c.set_transit_key(b"\x00")
+        new_handshake = c._build_relay_handshake()
+        token, old_handshake = self.old_build_relay_handshake(b"\x00")
+
+        tc = transit_server.TransitConnection()
+        tc.factory = mock.Mock()
+        tc.factory.connection_got_token = mock.Mock()
+        tc.dataReceived(new_handshake[:-1])
+        self.assertEqual(tc.factory.connection_got_token.mock_calls, [])
+        tc.dataReceived(new_handshake[-1:])
+        self.assertEqual(tc.factory.connection_got_token.mock_calls,
+                         [mock.call(hexlify(token), c._side.encode("ascii"), tc)])
+
+
+class Full(ServerBase, unittest.TestCase):
     def doBoth(self, d1, d2):
         return gatherResults([d1, d2], True)
 
     @inlineCallbacks
-    def test_full(self):
+    def test_direct(self):
         KEY = b"k"*32
         s = transit.TransitSender(None)
         r = transit.TransitReceiver(None)
@@ -1351,3 +1532,31 @@ class Full(unittest.TestCase):
 
         yield x.close()
         yield y.close()
+
+    @inlineCallbacks
+    def test_relay(self):
+        KEY = b"k"*32
+        s = transit.TransitSender(self.transit, no_listen=True)
+        r = transit.TransitReceiver(self.transit, no_listen=True)
+
+        s.set_transit_key(KEY)
+        r.set_transit_key(KEY)
+
+        shints = yield s.get_connection_hints()
+        rhints = yield r.get_connection_hints()
+
+        s.add_connection_hints(rhints)
+        r.add_connection_hints(shints)
+
+        (x,y) = yield self.doBoth(s.connect(), r.connect())
+        self.assertIsInstance(x, transit.Connection)
+        self.assertIsInstance(y, transit.Connection)
+
+        d = y.receive_record()
+
+        x.send_record(b"record1")
+        r = yield d
+        self.assertEqual(r, b"record1")
+
+        yield x.close()
+        yield y.close()
