@@ -522,6 +522,45 @@ class Basic(unittest.TestCase):
         self.assertEqual(len(pieces), 3) # nameplate plus two words
         self.assert_(re.search(r'^\d+-\w+-\w+$', code), code)
 
+    def _test_establish_key_hook(self, established, before):
+        timing = DebugTiming()
+        w = wormhole._Wormhole(APPID, "relay_url", reactor, None, timing)
+
+        if before:
+            d = w.establish_key()
+
+        if established is True:
+            w._key = b"key"
+        elif established is False:
+            w._key = None
+        else:
+            w._key = b"key"
+            w._error = WelcomeError()
+
+        if not before:
+            d = w.establish_key()
+        else:
+            w._maybe_notify_key()
+
+        if w._key is not None and established is True:
+            self.successResultOf(d)
+        elif established is False:
+            self.assertNot(d.called)
+        else:
+            self.failureResultOf(d)
+
+    def test_establish_key_hook(self):
+        for established in (True, False, "error"):
+            for before in (True, False):
+                self._test_establish_key_hook(established, before)
+
+    def test_establish_key_twice(self):
+        timing = DebugTiming()
+        w = wormhole._Wormhole(APPID, "relay_url", reactor, None, timing)
+        d = w.establish_key()
+        self.assertRaises(InternalError, w.establish_key)
+        del d
+
     # make sure verify() can be called both before and after the verifier is
     # computed
 
@@ -871,6 +910,40 @@ class Wormholes(ServerBase, unittest.TestCase):
         yield w1.close()
         yield w2.close()
 
+    @inlineCallbacks
+    def test_rx_dedup(self):
+        # Future clients will handle losing/reestablishing the Rendezvous
+        # Server connection by retransmitting messages, which will sometimes
+        # cause duplicate messages. Make sure this client can tolerate them.
+        # The first place this would fail was when the second copy of the
+        # incoming PAKE message was received, which would cause
+        # SPAKE2.finish() to be called a second time, which throws an error
+        # (which, being somewhat unexpected, caused a hang rather than a
+        # clear exception).
+        with mock.patch("wormhole.wormhole._Wormhole", MessageDoublingReceiver):
+            w1 = wormhole.wormhole(APPID, self.relayurl, reactor)
+        w2 = wormhole.wormhole(APPID, self.relayurl, reactor)
+        w1.set_code("123-purple-elephant")
+        w2.set_code("123-purple-elephant")
+        w1.send(b"data1"), w2.send(b"data2")
+        dl = yield self.doBoth(w1.get(), w2.get())
+        (dataX, dataY) = dl
+        self.assertEqual(dataX, b"data2")
+        self.assertEqual(dataY, b"data1")
+        yield w1.close()
+        yield w2.close()
+
+class MessageDoublingReceiver(wormhole._Wormhole):
+    # we could double messages on the sending side, but a future server will
+    # strip those duplicates, so to really exercise the receiver, we must
+    # double them on the inbound side instead
+    #def _msg_send(self, phase, body):
+    #    wormhole._Wormhole._msg_send(self, phase, body)
+    #    self._ws_send_command("add", phase=phase, body=bytes_to_hexstr(body))
+    def _event_received_peer_message(self, side, phase, body):
+        wormhole._Wormhole._event_received_peer_message(self, side, phase, body)
+        wormhole._Wormhole._event_received_peer_message(self, side, phase, body)
+
 class Errors(ServerBase, unittest.TestCase):
     @inlineCallbacks
     def test_codes_1(self):
