@@ -1,6 +1,6 @@
 # no unicode_literals, revisit after twisted patch
 from __future__ import print_function, absolute_import
-import re, sys, time, socket
+import os, re, sys, time, socket
 from collections import namedtuple, deque
 from binascii import hexlify, unhexlify
 import six
@@ -15,17 +15,12 @@ from nacl.secret import SecretBox
 from hkdf import Hkdf
 from .errors import InternalError
 from .timing import DebugTiming
+from .util import bytes_to_hexstr
 from . import ipaddrs
 
 def HKDF(skm, outlen, salt=None, CTXinfo=b""):
     return Hkdf(salt, skm).expand(CTXinfo, outlen)
 
-def debug(msg):
-    if False:
-        print(msg)
-def since(start):
-    return time.time() - start
-
 class TransitError(Exception):
     pass
 
@@ -76,9 +71,11 @@ def build_sender_handshake(key):
     hexid = HKDF(key, 32, CTXinfo=b"transit_sender")
     return b"transit sender "+hexlify(hexid)+b" ready\n\n"
 
-def build_relay_handshake(key):
+def build_sided_relay_handshake(key, side):
+    assert isinstance(side, type(u""))
+    assert len(side) == 8*2
     token = HKDF(key, 32, CTXinfo=b"transit_relay_token")
-    return b"please relay "+hexlify(token)+b"\n"
+    return b"please relay "+hexlify(token)+b" for side "+side.encode("ascii")+b"\n"
 
 
 # These namedtuples are "hint objects". The JSON-serializable dictionaries
@@ -90,11 +87,12 @@ def build_relay_handshake(key):
 # * expect to see the receiver/sender handshake bytes from the other side
 # * the sender writes "go\n", the receiver waits for "go\n"
 # * the rest of the connection contains transit data
-DirectTCPV1Hint = namedtuple("DirectTCPV1Hint", ["hostname", "port"])
-TorTCPV1Hint = namedtuple("TorTCPV1Hint", ["hostname", "port"])
-# RelayV1Hint contains a list of DirectTCPV1Hint and TorTCPV1Hint hints. For
-# each one, make the TCP connection, send the relay handshake, then complete
-# the rest of the V1 protocol. Only one hint per relay is useful.
+DirectTCPV1Hint = namedtuple("DirectTCPV1Hint", ["hostname", "port", "priority"])
+TorTCPV1Hint = namedtuple("TorTCPV1Hint", ["hostname", "port", "priority"])
+# RelayV1Hint contains a tuple of DirectTCPV1Hint and TorTCPV1Hint hints (we
+# use a tuple rather than a list so they'll be hashable into a set). For each
+# one, make the TCP connection, send the relay handshake, then complete the
+# rest of the V1 protocol. Only one hint per relay is useful.
 RelayV1Hint = namedtuple("RelayV1Hint", ["hints"])
 
 def describe_hint_obj(hint):
@@ -105,29 +103,40 @@ def describe_hint_obj(hint):
     else:
         return str(hint)
 
-def parse_hint_argv(hint):
+def parse_hint_argv(hint, stderr=sys.stderr):
     assert isinstance(hint, type(u""))
     # return tuple or None for an unparseable hint
+    priority = 0.0
     mo = re.search(r'^([a-zA-Z0-9]+):(.*)$', hint)
     if not mo:
-        print("unparseable hint '%s'" % (hint,))
+        print("unparseable hint '%s'" % (hint,), file=stderr)
         return None
     hint_type = mo.group(1)
     if hint_type != "tcp":
-        print("unknown hint type '%s' in '%s'" % (hint_type, hint))
+        print("unknown hint type '%s' in '%s'" % (hint_type, hint), file=stderr)
         return None
     hint_value = mo.group(2)
-    mo = re.search(r'^(.*):(\d+)$', hint_value)
+    pieces = hint_value.split(":")
+    if len(pieces) < 2:
+        print("unparseable TCP hint (need more colons) '%s'" % (hint,),
+              file=stderr)
+        return None
+    mo = re.search(r'^(\d+)$', pieces[1])
     if not mo:
-        print("unparseable TCP hint '%s'" % (hint,))
+        print("non-numeric port in TCP hint '%s'" % (hint,), file=stderr)
         return None
-    hint_host = mo.group(1)
-    try:
-        hint_port = int(mo.group(2))
-    except ValueError:
-        print("non-numeric port in TCP hint '%s'" % (hint,))
-        return None
-    return DirectTCPV1Hint(hint_host, hint_port)
+    hint_host = pieces[0]
+    hint_port = int(pieces[1])
+    for more in pieces[2:]:
+        if more.startswith("priority="):
+            more_pieces = more.split("=")
+            try:
+                priority = float(more_pieces[1])
+            except ValueError:
+                print("non-float priority= in TCP hint '%s'" % (hint,),
+                      file=stderr)
+                return None
+    return DirectTCPV1Hint(hint_host, hint_port, priority)
 
 TIMEOUT=15
 
@@ -150,7 +159,6 @@ class Connection(protocol.Protocol, policies.TimeoutMixin):
         self._waiting_reads = deque()
 
     def connectionMade(self):
-        debug("handle %r" %  (self.transport,))
         self.setTimeout(TIMEOUT) # does timeoutConnection() when it expires
         self.factory.connectionWasMade(self)
 
@@ -575,15 +583,18 @@ class Common:
 
     def __init__(self, transit_relay, no_listen=False, tor_manager=None,
                  reactor=reactor, timing=None):
+        self._side = bytes_to_hexstr(os.urandom(8)) # unicode
         if transit_relay:
             if not isinstance(transit_relay, type(u"")):
                 raise InternalError
-            relay = RelayV1Hint(hints=[parse_hint_argv(transit_relay)])
+            # TODO: allow multiple hints for a single relay
+            relay_hint = parse_hint_argv(transit_relay)
+            relay = RelayV1Hint(hints=(relay_hint,))
             self._transit_relays = [relay]
         else:
             self._transit_relays = []
         self._their_direct_hints = [] # hintobjs
-        self._their_relay_hints = []
+        self._our_relay_hints = set(self._transit_relays)
         self._tor_manager = tor_manager
         self._transit_key = None
         self._no_listen = no_listen
@@ -604,7 +615,7 @@ class Common:
             # some test hosts, including the appveyor VMs, *only* have
             # 127.0.0.1, and the tests will hang badly if we remove it.
             addresses = non_loopback_addresses
-        direct_hints = [DirectTCPV1Hint(six.u(addr), portnum)
+        direct_hints = [DirectTCPV1Hint(six.u(addr), portnum, 0.0)
                         for addr in addresses]
         ep = endpoints.serverFromString(reactor, "tcp:%d" % portnum)
         return direct_hints, ep
@@ -620,6 +631,7 @@ class Common:
         direct_hints = yield self._get_direct_hints()
         for dh in direct_hints:
             hints.append({u"type": u"direct-tcp-v1",
+                          u"priority": dh.priority,
                           u"hostname": dh.hostname,
                           u"port": dh.port, # integer
                           })
@@ -627,6 +639,7 @@ class Common:
             rhint = {u"type": u"relay-v1", u"hints": []}
             for rh in relay.hints:
                 rhint[u"hints"].append({u"type": u"direct-tcp-v1",
+                                        u"priority": rh.priority,
                                         u"hostname": rh.hostname,
                                         u"port": rh.port})
             hints.append(rhint)
@@ -686,10 +699,11 @@ class Common:
         if not(u"port" in hint and isinstance(hint[u"port"], int)):
             log.msg("invalid port in hint: %r" % (hint,))
             return None
+        priority = hint.get(u"priority", 0.0)
         if hint_type == u"direct-tcp-v1":
-            return DirectTCPV1Hint(hint[u"hostname"], hint[u"port"])
+            return DirectTCPV1Hint(hint[u"hostname"], hint[u"port"], priority)
         else:
-            return TorTCPV1Hint(hint[u"hostname"], hint[u"port"])
+            return TorTCPV1Hint(hint[u"hostname"], hint[u"port"], priority)
 
     def add_connection_hints(self, hints):
         for h in hints: # hint structs
@@ -703,10 +717,14 @@ class Common:
                 # with a set of equally-valid ways to connect to it. Treat
                 # them as separate relays, instead of merging them all
                 # together like this.
+                relay_hints = []
                 for rhs in h.get(u"hints", []):
-                    rh = self._parse_tcp_v1_hint(rhs)
-                    if rh:
-                        self._their_relay_hints.append(rh)
+                    h = self._parse_tcp_v1_hint(rhs)
+                    if h:
+                        relay_hints.append(h)
+                if relay_hints:
+                    rh = RelayV1Hint(hints=tuple(sorted(relay_hints)))
+                    self._our_relay_hints.add(rh)
             else:
                 log.msg("unknown hint type: %r" % (h,))
 
@@ -797,21 +815,32 @@ class Common:
             contenders.append(d)
             relay_delay = self.RELAY_DELAY
 
-        # Start trying the relay a few seconds after we start to try the
+        # Start trying the relays a few seconds after we start to try the
         # direct hints. The idea is to prefer direct connections, but not be
-        # afraid of using the relay when we have direct hints that don't
+        # afraid of using a relay when we have direct hints that don't
         # resolve quickly. Many direct hints will be to unused local-network
         # IP addresses, which won't answer, and would take the full TCP
         # timeout (30s or more) to fail.
-        for hint_obj in self._their_relay_hints:
-            ep = self._endpoint_from_hint_obj(hint_obj)
-            if not ep:
-                continue
-            description = "->relay:%s" % describe_hint_obj(hint_obj)
-            d = task.deferLater(self._reactor, relay_delay,
-                                self._start_connector, ep, description,
-                                is_relay=True)
-            contenders.append(d)
+
+        prioritized_relays = {}
+        for rh in self._our_relay_hints:
+            for hint_obj in rh.hints:
+                priority = hint_obj.priority
+                if priority not in prioritized_relays:
+                    prioritized_relays[priority] = set()
+                prioritized_relays[priority].add(hint_obj)
+
+        for priority in sorted(prioritized_relays, reverse=True):
+            for hint_obj in prioritized_relays[priority]:
+                ep = self._endpoint_from_hint_obj(hint_obj)
+                if not ep:
+                    continue
+                description = "->relay:%s" % describe_hint_obj(hint_obj)
+                d = task.deferLater(self._reactor, relay_delay,
+                                    self._start_connector, ep, description,
+                                    is_relay=True)
+                contenders.append(d)
+            relay_delay += self.RELAY_DELAY
 
         if not contenders:
             raise TransitError("No contenders for connection")
@@ -830,11 +859,14 @@ class Common:
         d.addBoth(_done)
         return d
 
+    def _build_relay_handshake(self):
+        return build_sided_relay_handshake(self._transit_key, self._side)
+
     def _start_connector(self, ep, description, is_relay=False):
         relay_handshake = None
         if is_relay:
             assert self._transit_key
-            relay_handshake = build_relay_handshake(self._transit_key)
+            relay_handshake = self._build_relay_handshake()
         f = OutboundConnectionFactory(self, relay_handshake, description)
         d = ep.connect(f)
         # fires with protocol, or ConnectError
