@@ -237,6 +237,8 @@ class _Wormhole:
         self._flag_need_to_see_mailbox_used = True
         self._flag_need_to_build_msg1 = True
         self._flag_need_to_send_PAKE = True
+        self._establish_key_called = False
+        self._key_waiter = None
         self._key = None
 
         self._version_message = None
@@ -283,6 +285,14 @@ class _Wormhole:
 
     # todo: restore-saved-state entry points
 
+    def establish_key(self):
+        """
+        returns a Deferred that fires when we've established the shared key.
+        When successful, the Deferred fires with a simple `True`, otherwise
+        it fails.
+        """
+        return self._API_establish_key()
+
     def verify(self):
         """Returns a Deferred that fires when we've heard back from the other
         side, and have confirmed that they used the right wormhole code. When
@@ -559,6 +569,7 @@ class _Wormhole:
 
     def _event_established_key(self):
         self._timing.add("key established")
+        self._maybe_notify_key()
 
         # both sides send different (random) version messages
         self._send_version_message()
@@ -569,6 +580,25 @@ class _Wormhole:
         self._maybe_check_version()
         self._maybe_send_phase_messages()
 
+    def _API_establish_key(self):
+        if self._error: return defer.fail(self._error)
+        if self._establish_key_called: raise InternalError
+        self._establish_key_called = True
+        if self._key is not None:
+            return defer.succeed(True)
+        self._key_waiter = defer.Deferred()
+        return self._key_waiter
+
+    def _maybe_notify_key(self):
+        if self._key is None:
+            return
+        if self._error:
+            result = failure.Failure(self._error)
+        else:
+            result = True
+        if self._key_waiter and not self._key_waiter.called:
+            self._key_waiter.callback(result)
+
     def _send_version_message(self):
         # this is encrypted like a normal phase message, and includes a
         # dictionary of version flags to let the other Wormhole know what
@@ -728,10 +758,15 @@ class _Wormhole:
 
         if self._closing:
             log.msg("received peer message while closing '%s'" % phase)
+        if phase in self._received_messages:
+            log.msg("ignoring duplicate peer message '%s'" % phase)
+            return
 
         if phase == "pake":
+            self._received_messages["pake"] = body
             return self._event_received_pake(body)
         if phase == "version":
+            self._received_messages["version"] = body
             return self._event_received_version(side, body)
         if re.search(r'^\d+$', phase):
             return self._event_received_phase_message(side, phase, body)
@@ -824,6 +859,9 @@ class _Wormhole:
         if self._verifier_waiter and not self._verifier_waiter.called:
             if self.DEBUG: print("EB VW")
             self._verifier_waiter.errback(error)
+        if self._key_waiter and not self._key_waiter.called:
+            if self.DEBUG: print("EB KW")
+            self._key_waiter.errback(error)
         for d in self._receive_waiters.values():
             if self.DEBUG: print("EB RW")
             d.errback(error)
